@@ -4,44 +4,49 @@
 # nodeformat.py, provides non-GUI base classes for node formating info
 #
 # TreeLine, an information storage program
-# Copyright (C) 2005, Douglas W. Bell
+# Copyright (C) 2006, Douglas W. Bell
 #
 # This is free software; you can redistribute it and/or modify it under the
-# terms of the GNU General Public License, Version 2.  This program is
-# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
+# terms of the GNU General Public License, either Version 2 or any later
+# version.  This program is distributed in the hope that it will be useful,
+# but WITTHOUT ANY WARRANTY.  See the included LICENSE file for details.
 #****************************************************************************
 
-import re, copy, types, sys, os.path, stat, time
+import re
+import copy
+import types
+import sys
+import os.path
+import stat
+import time
 from xml.sax.saxutils import escape
-if sys.platform != 'win32':
+if not sys.platform.startswith('win'):
     import pwd
-from fieldformat import TextFormat, LongTextFormat, NumberFormat, \
-                        ChoiceFormat, CombinationFormat, AutoChoiceFormat, \
-                        DateFormat, TimeFormat, BooleanFormat, URLFormat, \
-                        PathFormat, EmailFormat, InternalLinkFormat, \
-                        ExecuteLinkFormat, PictureFormat, ParentFormat, \
-                        AncestorFormat, ChildFormat, xslEscape
-from gendate import GenDate
-from gentime import GenTime
-from conditional import Conditional
+import fieldformat
+import conditional
+import gendate
+import gentime
 import treedoc
 import globalref
 
 
-class NodeFormat:
+class NodeFormat(object):
     """Stores node type field list and line formatting"""
     lineAttrRe = re.compile('line\d+$')
-    fieldSplitRe = re.compile(r'({\*(?:\**|\?|!|&)[\w_\-.]+\*})', re.U)
-    fieldPartRe = re.compile(r'{\*(\**|\?|!|&)([\w_\-.]+)\*}', re.U)
+    fieldSplitRe = re.compile(r'({\*(?:\**|\?|!|&|#)[\w_\-.]+\*})', re.U)
+    fieldPartRe = re.compile(r'{\*(\**|\?|!|&|#)([\w_\-.]+)\*}', re.U)
+    levelFieldRe = re.compile(r'[^0-9]+([0-9]+)$')
     def __init__(self, name, formatAttrs={}, defaultField=''):
         self.name = name
         self.fieldList = []
         self.lineList = []
         self.numEmptyFields = 0
         self.numFullFields = 0
+        self.uniqueIDFields = []
         self.childType = formatAttrs.get(u'childtype', '')
         self.genericType = formatAttrs.get(u'generic', '')
-        self.conditional = Conditional(formatAttrs.get(u'condition', ''))
+        self.conditional = conditional.Conditional(formatAttrs.
+                                                   get(u'condition', ''))
         self.sibPrefix = formatAttrs.get(u'sibprefix', '')
         self.sibSuffix = formatAttrs.get(u'sibsuffix', '')
         self.iconName = formatAttrs.get(u'icon', '')
@@ -81,7 +86,7 @@ class NodeFormat:
         """Return list of first fieldname in each line"""
         result = []
         for line in self.lineList[1:]:
-            strippedLine = [part for part in line \
+            strippedLine = [part for part in line
                             if type(part) not in types.StringTypes]
             result.append(strippedLine and strippedLine[0].name or '')
         return result
@@ -95,7 +100,8 @@ class NodeFormat:
 
     def addNewField(self, name, attrs={}):
         """Add new field with type, format, extra text in attrs dict"""
-        field = globals()[attrs.get(u'type', 'Text') + 'Format'](name, attrs)
+        typeName = '%sFormat' % attrs.get(u'type', 'Text')
+        field = fieldformat.__dict__[typeName](name, attrs)
         self.fieldList.append(field)
         if attrs.get(u'ref', '').startswith('y'):
             self.refField = field
@@ -103,18 +109,24 @@ class NodeFormat:
             self.refField = self.fieldList[0]
         sepName = field.sepName()
         # replace field name with the field
-        self.lineList = [[part == sepName and field or part for part in line] \
+        self.lineList = [[part == sepName and field or part for part in line]
                          for line in self.lineList]
 
     def addFieldIfNew(self, name, attrs={}):
-        """"Add new field if it doesn't exist"""
+        """Add new field if it doesn't exist,
+           add to generic for a derived type"""
         if name not in self.fieldNames():
             self.addNewField(name, attrs)
             if self.genericType:
-                genericType = globalref.docRef.treeFormats.\
-                                        findFormat(self.genericType)
-                if genericType:
-                    genericType.addNewField(name, attrs)
+                try:
+                    genericType = globalref.docRef.\
+                                  treeFormats[self.genericType]
+                except KeyError:
+                    print 'Warning - generic type %s not found' % \
+                         self.genericType.encode(globalref.localTextEncoding)
+                    self.genericType = ''
+                    return
+                genericType.addNewField(name, attrs)
 
     def addLine(self, text):
         """Add format line to end of lineList, line 0 is the title"""
@@ -135,6 +147,14 @@ class NodeFormat:
         """Change the title format line"""
         self.insertLine(text, 0)
 
+    def changeOutputLines(self, lines):
+        """Replace the output format lines with given list"""
+        self.lineList = self.lineList[:1]
+        if not self.lineList:
+            self.lineList = ['']
+        for line in lines:
+            self.addLine(line)
+
     def updateLineFields(self):
         """Re-find fields to update for any changes in the fieldList"""
         self.lineList = [self.parseLine(line) for line in self.getLines()]
@@ -153,15 +173,27 @@ class NodeFormat:
                     field = self.findField(fieldName)
                     newLine.append(field and field or part)
                 elif modifier[0] == '*':
-                    newLine.append(ParentFormat(fieldName, len(modifier)))
+                    newLine.append(fieldformat.ParentFormat(fieldName,
+                                                            len(modifier)))
                 elif modifier == '?':
-                    newLine.append(AncestorFormat(fieldName))
+                    newLine.append(fieldformat.AncestorFormat(fieldName))
                 elif modifier == '&':
-                    newLine.append(ChildFormat(fieldName))
-                else:  # == '!' file field
-                    field = globalref.docRef.fileInfoItem.nodeFormat.\
+                    newLine.append(fieldformat.ChildFormat(fieldName))
+                elif modifier == '!':     # file field
+                    field = globalref.docRef.fileInfoFormat.\
                                findField(fieldName)
                     newLine.append(field and field or part)
+                elif modifier == '#':
+                    match = NodeFormat.levelFieldRe.match(fieldName)
+                    if match:
+                        level = int(match.group(1))
+                        if level > 0:
+                            newLine.append(fieldformat.CountFormat(fieldName,
+                                                                   level))
+                    else:
+                        newLine.append(part)
+                else:
+                    newLine.append(part)
             else:
                 newLine.append(part)
         return newLine
@@ -179,11 +211,11 @@ class NodeFormat:
         """Return string with formatted title data"""
         if not self.lineList:
             return ''
-        line = u''.join([self.fieldText(text, item, True) for text in \
+        line = u''.join([self.fieldText(text, item, True) for text in
                          self.lineList[0]])
-        return line.strip()
+        return line.strip().split('\n', 1)[0]  # truncate to 1st line
 
-    def formatText(self, item, skipEmpty=True, addPrefix=False, \
+    def formatText(self, item, skipEmpty=True, addPrefix=False,
                    addSuffix=False, internal=False):
         """Return list of output strings from formatting data"""
         tagExp = re.compile('.*(<br[ /]*?>|<BR[ /]*?>|<hr[ /]*?>|<HR[ /]*?>)$')
@@ -191,7 +223,7 @@ class NodeFormat:
         for lineData in self.lineList[1:]:
             self.numEmptyFields = 0
             self.numFullFields = 0
-            line = u''.join([self.fieldText(text, item, False, internal) \
+            line = u''.join([self.fieldText(text, item, False, internal)
                              for text in lineData])
             if self.numFullFields or not self.numEmptyFields or not skipEmpty:
                 result.append(line.strip())   # add if tags are not empty
@@ -219,7 +251,7 @@ class NodeFormat:
         for lineData in self.lineList:
             self.numEmptyFields = 0
             self.numFullFields = 0
-            line = u''.join([self.fieldText(text, item, False) for text \
+            line = u''.join([self.fieldText(text, item, False) for text
                              in lineData])
             if self.numFullFields or not self.numEmptyFields:
                 result.append(line.strip())   # add if tags are not empty
@@ -227,12 +259,29 @@ class NodeFormat:
                 result.append('')
         return result
 
-    def fieldText(self, field, item, titleMode=False, internal=False):
+    def formatPlainTextLines(self, item):
+        """Return a list of all text lines (title & output), using
+           plain text (titleMode) for all"""
+        result = []
+        for lineData in self.lineList:
+            self.numEmptyFields = 0
+            self.numFullFields = 0
+            line = u''.join([self.fieldText(text, item, True, False, True)
+                             for text in lineData])
+            if not result or (line and \
+                              (self.numFullFields or not self.numEmptyFields)):
+                result.append(line.strip())   # add if tags are not empty
+        return result
+
+    def fieldText(self, field, item, titleMode=False, internal=False,
+                  stripFormatTags=False):
         """Return formatted text for field"""
         if type(field) in types.StringTypes:
             text = field
             if not titleMode and not globalref.docRef.formHtml:
                 text = escape(text)
+            if stripFormatTags and globalref.docRef.formHtml:
+                text = fieldformat.TextFormat.stripTagRe.sub('', text)
         else:
             text = field.outputText(item, titleMode, internal)
             if text:
@@ -252,7 +301,7 @@ class NodeFormat:
                 pattern += re.escape(seg)
                 extraText += seg
             elif seg.parentLevel:
-                text = self.fieldText(seg, item, True)
+                text = self.fieldText(seg, item, True).split('\n', 1)[0]
                 pattern += re.escape(text)
                 extraText += text
             else:
@@ -271,8 +320,8 @@ class NodeFormat:
             for field in fields[1:]:
                 item.data[field.name] = ''
         if addUndo and globalref.pluginInterface:
-            globalref.pluginInterface.execCallback(globalref.pluginInterface.\
-                                                   dataChangeCallbacks, item, \
+            globalref.pluginInterface.execCallback(globalref.pluginInterface.
+                                                   dataChangeCallbacks, item,
                                                    fields)
         return True
 
@@ -289,14 +338,14 @@ class NodeFormat:
         if self.genericType:
             result.append(u'generic="%s"' % self.genericType)
         if self.conditional:
-            result.append(u'condition="%s"' % escape(self.conditional.\
-                                                     conditionText(), \
+            result.append(u'condition="%s"' % escape(self.conditional.
+                                                     conditionText(),
                                                      treedoc.escDict))
         if self.sibPrefix:
-            result.append(u'sibprefix="%s"' % escape(self.sibPrefix, \
+            result.append(u'sibprefix="%s"' % escape(self.sibPrefix,
                                                      treedoc.escDict))
         if self.sibSuffix:
-            result.append(u'sibsuffix="%s"' % escape(self.sibSuffix, \
+            result.append(u'sibsuffix="%s"' % escape(self.sibSuffix,
                                                      treedoc.escDict))
         if self.iconName:
             result.append(u'icon="%s"' % self.iconName)
@@ -304,8 +353,8 @@ class NodeFormat:
 
     def getLines(self, englishOnly=False):
         """Return text list of formatting lines"""
-        lines = [u''.join([self.fieldName(part, englishOnly) \
-                           for part in line]) \
+        lines = [u''.join([self.fieldName(part, englishOnly)
+                           for part in line])
                  for line in self.lineList]
         return lines and lines or ['']
 
@@ -334,44 +383,44 @@ class NodeFormat:
         output = [u'', u'<xsl:template match="%s[@item=\'y\']">' % self.name]
         if addAnchors:
             name = self.refField.name
-            output.extend([u'<xsl:if test="normalize-space(./%s)">' % name, \
-                           u'<xsl:for-each select="./%s">' % name, \
+            output.extend([u'<xsl:if test="normalize-space(./%s)">' % name,
+                           u'<xsl:for-each select="./%s">' % name,
                            u'<a id="{.}" />', u'</xsl:for-each></xsl:if>'])
         if self.sibPrefix:
-            output.append(xslEscape(self.sibPrefix))
+            output.append(fieldformat.xslEscape(self.sibPrefix))
         for lineDataRaw in lineList:
-            lineData = [(type(field) not in types.StringTypes) and field or \
-                        hrFormat.join(hrExp.split(brFormat.\
-                                                  join(brExp.split(field)))) \
+            lineData = [(type(field) not in types.StringTypes) and field or
+                        hrFormat.join(hrExp.split(brFormat.
+                                                  join(brExp.split(field))))
                         for field in lineDataRaw]
-            fields = [field for field in lineData if \
+            fields = [field for field in lineData if
                       type(field) not in types.StringTypes]
             if fields:
                 text = u'<xsl:if test="%s">' % \
                        u' or '.join([field.xslTestText() for field in fields])
                 output.append(text)
-                tagMatch = tagExp.match(type(lineData[-1]) in \
-                                        types.StringTypes and \
+                tagMatch = tagExp.match(type(lineData[-1]) in
+                                        types.StringTypes and
                                         lineData[-1] or '')
                 if tagMatch:
                     if len(tagMatch.group(1)) == len(lineData[-1]):
                         del lineData[-1]
                     else:
                         lineData[-1] = lineData[-1][:-len(tagMatch.group(1))]
-                line = u''.join([(type(text) in types.StringTypes) \
-                                 and xslEscape(text) or text.xslText() \
-                                 for text in lineData])
+                line = u''.join([(type(text) in types.StringTypes) and
+                                 fieldformat.xslEscape(text) or
+                                 text.xslText() for text in lineData])
                 output.extend([line + brFormat, u'</xsl:if>'])
                 if tagMatch:
                     output.append(tagMatch.group(1))
             else:
                 output.append(lineData[0] + brFormat)
         if self.sibSuffix:
-            output.append(xslEscape(self.sibSuffix))
+            output.append(fieldformat.xslEscape(self.sibSuffix))
         if globalref.docRef.spaceBetween:
             output.append(u'<br/>')
-        output.extend([u'<div style="margin-left:%ipx">' % indent, \
-                       u'<xsl:apply-templates/>', u'</div>', \
+        output.extend([u'<div style="margin-left:%ipx">' % indent,
+                       u'<xsl:apply-templates/>', u'</div>',
                        u'</xsl:template>'])
         return output
 
@@ -420,7 +469,7 @@ class NodeFormat:
             return availFields[0]
         if not availFields:
             return None
-        for srchTerm in [globalref.docRef.linkFieldName, u'link', u'url', \
+        for srchTerm in [globalref.docRef.linkFieldName, u'link', u'url',
                          u'href']:
             for field in availFields:
                 name = field.name.lower()
@@ -435,21 +484,29 @@ class NodeFormat:
         """Return a list of fields that need to have choices added"""
         return [field for field in self.fieldList if field.autoAddChoices]
 
-    def setInitDefaultData(self, data):
+    def findUniqueIDFields(self):
+        """Set memeber variable and return list of unique ID fields"""
+        self.uniqueIDFields = [field for field in self.fieldList
+                               if field.typeName == 'UniqueID']
+        return self.uniqueIDFields
+
+    def setInitDefaultData(self, data, onlyIfBlank=False):
         """Add initial default data from fields into supplied dict"""
         for field in self.fieldList:
             text = field.getInitDefault()
-            if text:
+            if text and (not onlyIfBlank or not data.get(field.name, '')):
                 data[field.name] = text
 
-    def updateFromGeneric(self, treeFormats):
+    def updateFromGeneric(self):
         """Update the field names and types in self (a derived type)
            to match generic"""
         if not self.genericType:
             return
-        genericType = treeFormats.findFormat(self.genericType)
-        if not genericType:
-            print 'Warning - generic type %s not found' % self.genericType
+        try:
+            genericType = globalref.docRef.treeFormats[self.genericType]
+        except KeyError:
+            print 'Warning - generic type %s not found' % \
+                  self.genericType.encode(globalref.localTextEncoding)
             self.genericType = ''
             return
         newFields = []
@@ -483,7 +540,7 @@ class FileInfoFormat(NodeFormat):
         self.addNewField(FileInfoFormat.sizeFieldName, {'type': 'Number'})
         self.addNewField(FileInfoFormat.dateFieldName, {'type': 'Date'})
         self.addNewField(FileInfoFormat.timeFieldName, {'type': 'Time'})
-        if sys.platform != 'win32':
+        if not sys.platform.startswith('win'):
             self.addNewField(FileInfoFormat.ownerFieldName)
         # page info only for print header:
         self.addNewField(FileInfoFormat.pageNumFieldName)
@@ -506,9 +563,11 @@ class FileInfoFormat(NodeFormat):
         item.data[_(FileInfoFormat.pathFieldName)] = os.path.dirname(fileName)
         item.data[_(FileInfoFormat.sizeFieldName)] = str(status[stat.ST_SIZE])
         modTime = time.localtime(status[stat.ST_MTIME])
-        item.data[_(FileInfoFormat.dateFieldName)] = repr(GenDate(modTime))
-        item.data[_(FileInfoFormat.timeFieldName)] = repr(GenTime(modTime))
-        if sys.platform != 'win32':
+        item.data[_(FileInfoFormat.dateFieldName)] = \
+                        repr(gendate.GenDate(modTime))
+        item.data[_(FileInfoFormat.timeFieldName)] = \
+                        repr(gentime.GenTime(modTime))
+        if not sys.platform.startswith('win'):
             try:
                 item.data[_(FileInfoFormat.ownerFieldName)] = \
                                    pwd.getpwuid(status[stat.ST_UID])[0]
@@ -539,15 +598,14 @@ class FileInfoFormat(NodeFormat):
     def replaceListFormat(self):
         """Copy settings from the treeFormats list to here and
            replace the list's file format object"""
-        listFormat = globalref.docRef.treeFormats.findFormat(self.name)
+        listFormat = globalref.docRef.treeFormats.get(self.name, None)
         if listFormat:
-            for thisField, listField in zip(self.fieldList, \
+            for thisField, listField in zip(self.fieldList,
                                             listFormat.fieldList):
                 thisField.duplicateSettings(listField)
                 thisField.useFileInfo = True
             self.lineList = listFormat.lineList
-            globalref.docRef.treeFormats.remove(listFormat)
-            globalref.docRef.treeFormats.append(self)
+            globalref.docRef.treeFormats[self.name] = self
 
     def translateFields(self):
         """Translate field names into the current language,
@@ -557,3 +615,15 @@ class FileInfoFormat(NodeFormat):
             if transName != field.name:
                 field.enName = field.name
                 field.name = transName
+
+
+class ChildCountFormat(NodeFormat):
+    """Placeholder format for child count fields,
+       should not show up in main format type list"""
+    countFieldName = _('Level', 'child count field')
+    def __init__(self):
+        NodeFormat.__init__(self, u'CountFormat')
+        for level in range(3):
+            name = '%s%d' % (ChildCountFormat.countFieldName, level + 1)
+            field = fieldformat.CountFormat(name, level + 1)
+            self.fieldList.append(field)
